#
# Copyright (C) 2015  Ilya Markov
#
# Full copyright notice can be found in LICENSE.
#
import russir2016_utils

import sys
from enum import Enum
from collections import OrderedDict

from pyclick.click_models.ClickModel import ClickModel
from pyclick.click_models.Inference import Inference, EMInference
from pyclick.click_models.Param import ParamEM
from pyclick.click_models.ParamContainer import QueryDocumentParamContainer, RankParamContainer, SingleParamContainer
from pyclick.search_session.SearchSession import SearchSession
from pyclick.click_models.Evaluation import LogLikelihood, Perplexity


class TCMUtils(object):
    @staticmethod
    def is_last_session(search_task, search_session):
        session_index = search_task.search_sessions.index(search_session)
        return session_index == len(search_task.search_sessions) - 1
    
    @staticmethod
    def get_previous_results(search_task, search_session):
        previous_results = set()
        
        session_index = search_task.search_sessions.index(search_session)
        for i in range(session_index):
            for result in search_task.search_sessions[i].web_results:
                previous_results.add(result.id)
                
        return previous_results
    
    @staticmethod
    def get_no_clicks_given_match_prob(search_session, session_params, previous_results):
        no_clicks_prob = 1.0

        for rank, result in enumerate(search_session.web_results):
            attr = session_params[rank][TCM.param_names.attr].value()
            exam = session_params[rank][TCM.param_names.exam].value()
            fresh = session_params[rank][TCM.param_names.fresh].value() \
                if (result.id in previous_results) else 1.0

            no_clicks_prob *= (1 - attr * exam * fresh)

        return no_clicks_prob

    @staticmethod
    def get_match_given_session_prob(search_session, session_params, previous_results, is_last_session):
        if any(search_session.get_clicks()) or is_last_session:
            return 1.0
        else:
            match = session_params[0][TCM.param_names.match].value()
            new = session_params[0][TCM.param_names.new].value()
            p = TCMUtils.get_no_clicks_given_match_prob(search_session, session_params,
                                                          previous_results)
            # Here we simplify the computation assuming that the next sessions' clicks
            # do not depend on the current session given the value of N parameter.
            return 1.0 / (1 + (1.0 / match - 1) / (p * new))
            
            
class EMInference(Inference):
    ITERATION_NUM = 50

    def infer_params(self, click_model, search_sessions):
        if search_sessions is None or len(search_sessions) == 0:
            return

        for iteration in xrange(self.ITERATION_NUM):
            new_click_model = click_model.__class__()

            for search_session in search_sessions:
                current_session_params = click_model.get_session_params(search_session)
                new_session_params = new_click_model.get_session_params(search_session)

                for rank, result in enumerate(search_session.web_results):
                    for param_name, param in new_session_params[rank].items():
                        param.update(search_session, rank, current_session_params)

            click_model.params = new_click_model.params
            
            
class TCMAttr(ParamEM):
    def update(self, search_session, rank, session_params):
        """
        Updates the value of the parameter based on the observed search session.
        Note that this parameter corresponds to a search result at the given rank.

        :param search_session: The observed search session.
        :param rank: The rank of a search result, which corresponds to the current parameter.
        """
        attr = session_params[rank][TCM.param_names.attr].value()
        exam = session_params[rank][TCM.param_names.exam].value()
        
        if search_session.web_results[rank].click:
            self._numerator += 1
        else:
            self._numerator += (1 - exam) * attr / (1 - exam * attr)

        self._denominator += 1
        
        
class TCMExam(ParamEM):
    def update(self, search_session, rank, session_params):
        """
        Updates the value of the parameter based on the observed search session.
        Note that this parameter corresponds to a search result at the given rank.

        :param search_session: The observed search session.
        :param rank: The rank of a search result, which corresponds to the current parameter.
        """
        attr = session_params[rank][TCM.param_names.attr].value()
        exam = session_params[rank][TCM.param_names.exam].value()
        
        if search_session.web_results[rank].click:
            self._numerator += 1
        else:
            self._numerator += (1 - attr) * exam / (1 - exam * attr)

        self._denominator += 1
        
        
class TCM(ClickModel):
    param_names = Enum('TCMParamNames', 'attr exam')
    """The names of the TCM parameters."""

    def __init__(self):
        self.params = {self.param_names.attr: QueryDocumentParamContainer(TCMAttr),
                       self.param_names.exam: RankParamContainer.default(TCMExam)}
        self._inference = EMInference()

    def get_full_click_probs(self, search_session):
        """
        Returns a list of full click probabilities P(C = 1) for all results in the given search session.
        """
        click_probs = []
        
        session_params = self.get_session_params(search_session)
        for session_param in session_params:
            attr = session_param[self.param_names.attr].value()
            exam = session_param[self.param_names.exam].value()

            click_prob = attr * exam
            click_probs.append(click_prob)
        
        return click_probs
        
    def get_conditional_click_probs(self, search_session):
        """
        Returns a list of click probabilities conditioned
        on the observed clicks in the given search session.
        In particular, for a result at rank k calculates the following probability:
        P(C_k | C_1, C_2, ..., C_k-1),
        where C_i is 1 if there is a click on the i-th result
        in the given search session and 0 otherwise.
        """
        click_probs = self.get_full_click_probs(search_session)
        
        for rank, result in enumerate(search_session.web_results):
            if not result.click:
                click_probs[rank] = 1 - click_probs[rank]
        
        return click_probs
        
        
class TaskCentricSearchSession(SearchSession):
    def __init__(self, task, query):
        super(TaskCentricSearchSession, self).__init__(query)
        self.task = task
        
        
class SearchTask(object):
    def __init__(self, task):
        self._task = task
        self.search_sessions = []
        
    @staticmethod
    def get_search_tasks(search_sessions):
        """
        Groups search sessions by task and returns the list of all tasks.

        :param search_sessions: Task-centric search sessions.
        :returns: The list of tasks.
        """
        search_tasks = OrderedDict()

        for search_session in search_sessions:
            if search_session.task not in search_tasks:
                search_tasks[search_session.task] = SearchTask(search_session.task)

            search_tasks[search_session.task].search_sessions.append(search_session)

        return search_tasks.values()
        
        
if __name__ == "__main__":
    if len(sys.argv) < 3:
        print "USAGE: %s <dataset> <sessions_max>" % sys.argv[0]
        print "\tdataset - the path to the dataset from Yandex Relevance Prediction Challenge"
        print "\tsessions_max - the maximum number of one-query search sessions to consider"
        print ""
        sys.exit(1)

    search_sessions_path = sys.argv[1]
    search_sessions_num = int(sys.argv[2])
    train_sessions, train_queries, test_sessions, test_queries = russir2016_utils.get_train_test(
        search_sessions_path, search_sessions_num
    )

    print "Training on %d search sessions (%d unique queries)." % (len(train_sessions), len(train_queries))

    click_model = TCM()
    click_model.train(train_sessions)
    print click_model

    print "Testing on %d search sessions (%d unique queries)." % (len(test_sessions), len(test_queries))
        
    perplexity = Perplexity()
    print "Perplexity: %.4f" % perplexity.evaluate(click_model, test_sessions)[0]

    loglikelihood = LogLikelihood()
    print "Log-likelihood: %.4f" % loglikelihood.evaluate(click_model, test_sessions)
